# -*- coding: utf-8 -*-
#
#   Ming-Ke-Ming : Decentralized User Identity Authentication
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from typing import Optional

from .wrappers import StringWrapper

from .address import Address
from .address_impl import ANYWHERE, EVERYWHERE
from .identifier import ID, IDFactory


class Identifier(StringWrapper, ID):

    def __init__(self, identifier: str, name: Optional[str], address: Address, terminal: Optional[str]):
        super().__init__(string=identifier)
        self.__name = name
        self.__address = address
        self.__terminal = terminal

    @property  # Override
    def name(self) -> Optional[str]:
        return self.__name

    @property  # Override
    def address(self) -> Address:
        return self.__address

    @property  # Override
    def terminal(self) -> Optional[str]:
        return self.__terminal


class IdentifierFactory(IDFactory):

    def __init__(self):
        super().__init__()
        self.__ids = {}

    # Override
    def generate_identifier(self, meta, network: int, terminal: Optional[str]) -> ID:
        address = Address.generate(meta=meta, network=network)
        assert address is not None, 'failed to generate ID with meta: %s' % meta
        return ID.create(address=address, name=meta.seed, terminal=terminal)

    # Override
    def create_identifier(self, address: Address, name: Optional[str] = None, terminal: Optional[str] = None) -> ID:
        identifier = concat(address=address, name=name, terminal=terminal)
        cid = self.__ids.get(identifier)
        if cid is None:
            cid = Identifier(identifier=identifier, address=address, name=name, terminal=terminal)
            self.__ids[identifier] = cid
        return cid

    # Override
    def parse_identifier(self, identifier: str) -> Optional[ID]:
        cid = self.__ids.get(identifier)
        if cid is None:
            cid = parse(string=identifier)
            if cid is not None:
                self.__ids[identifier] = cid
        return cid


def parse(string: str) -> Optional[ID]:
    # split ID string
    pair = string.split('/', 1)
    # terminal
    if len(pair) == 1:
        # no terminal
        terminal = None
    else:
        # got terminal
        terminal = pair[1]
    # name @ address
    assert len(pair[0]) > 0, 'ID error: %s' % string
    pair = pair[0].split('@', 1)
    if len(pair) == 1:
        # got address without name
        name = None
        address = Address.parse(address=pair[0])
    else:
        # got name & address
        name = pair[0]
        address = Address.parse(address=pair[1])
    if address is not None:
        return Identifier(identifier=string, address=address, name=name, terminal=terminal)


def concat(address: Address, name: Optional[str] = None, terminal: Optional[str] = None) -> str:
    string = str(address)
    if name is not None and len(name) > 0:
        string = name + '@' + string
    if terminal is not None and len(terminal) > 0:
        string = string + '/' + terminal
    return string


# register ID factory
ID.register(factory=IdentifierFactory())


"""
    ID for Broadcast
    ~~~~~~~~~~~~~~~~
"""

ANYONE = ID.create(name='anyone', address=ANYWHERE)
EVERYONE = ID.create(name='everyone', address=EVERYWHERE)

# DIM Founder
FOUNDER = ID.create(name='moky', address=ANYWHERE)



# -*- coding: utf-8 -*-
#
#   Ming-Ke-Ming : Decentralized User Identity Authentication
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from .types import NetworkType, MetaType
from .address import Address, AddressFactory
from .address_impl import BaseAddressFactory
from .address_impl import ANYWHERE, EVERYWHERE
from .identifier import ID
from .id_impl import ANYONE, EVERYONE, FOUNDER
from .meta import Meta, MetaFactory
from .meta_impl import BaseMeta
from .tai_doc import Document, DocumentFactory
from .tai_docs import Visa, Bulletin
from .profile import BaseDocument, BaseVisa, BaseBulletin

name = "MingKeMing"

__author__ = 'Albert Moky'

__all__ = [

    'NetworkType', 'MetaType',
    'Address', 'AddressFactory', 'BaseAddressFactory',
    'ANYWHERE', 'EVERYWHERE',
    'ID', 'ANYONE', 'EVERYONE', 'FOUNDER',
    'Meta', 'BaseMeta',
    'MetaFactory',
    'DocumentFactory',
    'Document', 'BaseDocument',
    'Visa', 'BaseVisa',
    'Bulletin', 'BaseBulletin',
]



# -*- coding: utf-8 -*-
# ==============================================================================
# MIT License
#
# Copyright (c) 2020 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from abc import ABC, abstractmethod
from typing import Any, MutableMapping

from .string import StringWrapper


class Wrapper:
    """
        Wrapper for dict, list, str
    """

    @classmethod
    def unwrap(cls, o: Any, circularly: bool = False) -> Any:
        # unwrap string
        if isinstance(o, StringWrapper):
            return o.string
        if isinstance(o, str):
            return o
        # unwrap container
        if circularly:
            # unwrap map circularly
            if isinstance(o, MapWrapper):
                return MapWrapper.unwrap(o.dictionary, circularly=True)
            if isinstance(o, dict):
                return MapWrapper.unwrap(o, circularly=True)
            # unwrap list circularly
            if isinstance(o, ArrayWrapper):
                return ArrayWrapper.unwrap(o.array, circularly=True)
            if isinstance(o, list):
                return ArrayWrapper.unwrap(o, circularly=True)
        else:
            # unwrap map
            if isinstance(o, MapWrapper):
                return o.dictionary
            # unwrap list
            if isinstance(o, ArrayWrapper):
                return o.array
        # others
        return o


class ArrayWrapper(ABC):
    """
        List: Any
    """

    @classmethod
    def unwrap(cls, a: list, circularly: bool = False) -> list:
        # unwrap list container
        if isinstance(a, ArrayWrapper):
            a = a.array
        if not circularly:
            return a
        # unwrap items circularly
        t = []
        for o in a:
            item = Wrapper.unwrap(o, circularly=True)
            t.append(item)
        return t

    @property
    def array(self) -> list:
        raise NotImplemented

    @abstractmethod
    def copy_array(self, deep_copy: bool = False) -> list:
        raise NotImplemented


class MapWrapper(MutableMapping, ABC):
    """
        Mapping: str -> Any
    """

    @classmethod
    def unwrap(cls, d: dict, circularly: bool = False) -> dict:
        # unwrap map container
        if isinstance(d, MapWrapper):
            d = d.dictionary
        if not circularly:
            return d
        # unwrap keys, values circularly
        t = {}
        for k in d:
            v = d[k]
            key = StringWrapper.unwrap(k)
            value = Wrapper.unwrap(v, circularly=True)
            t[key] = value
        return t

    @property
    def dictionary(self) -> dict:
        raise NotImplemented

    @abstractmethod
    def copy_dictionary(self, deep_copy: bool = False) -> dict:
        raise NotImplemented



# -*- coding: utf-8 -*-
#
#   Ming-Ke-Ming : Decentralized User Identity Authentication
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from abc import ABC
from typing import Optional, Dict

from .wrappers import StringWrapper

from .types import NetworkType
from .address import Address, AddressFactory


"""
    Address for Broadcast
    ~~~~~~~~~~~~~~~~~~~~~
"""


class BroadcastAddress(StringWrapper, Address):

    def __init__(self, address: str, network: NetworkType):
        super().__init__(string=address)
        self.__network = network.value

    @property  # Override
    def network(self) -> int:
        return self.__network

    @property  # Override
    def is_broadcast(self) -> bool:
        return True


ANYWHERE = BroadcastAddress(address='anywhere', network=NetworkType.MAIN)
EVERYWHERE = BroadcastAddress(address='everywhere', network=NetworkType.GROUP)


"""
    Base Address Factory
    ~~~~~~~~~~~~~~~~~~~~
    
    abstractmethod:
        - create_address(address)
"""


class BaseAddressFactory(AddressFactory, ABC):

    def __init__(self):
        super().__init__()
        # cache broadcast addresses
        self._addresses: Dict[str, Address] = {
            str(ANYWHERE): ANYWHERE,
            str(EVERYWHERE): EVERYWHERE,
        }

    # Override
    def generate_address(self, meta, network: int) -> Optional[Address]:
        address = meta.generate_address(network=network)
        if address is not None:
            self._addresses[str(address)] = address
        return address

    # Override
    def parse_address(self, address: str) -> Optional[Address]:
        add = self._addresses.get(address)
        if add is None:
            add = Address.create(address=address)
            if add is not None:
                self._addresses[address] = add
        return add



# -*- coding: utf-8 -*-
#
#   Ming-Ke-Ming : Decentralized User Identity Authentication
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from abc import ABC, abstractmethod
from typing import Optional, Union, Any

from .wrappers import MapWrapper

from .crypto import utf8_encode
from .crypto import VerifyKey, SignKey

from .types import MetaType, meta_has_seed
from .address import Address
from .identifier import ID
from .factories import Factories


class Meta(MapWrapper, ABC):
    """This class is used to generate entity ID

        User/Group Meta data
        ~~~~~~~~~~~~~~~~~~~~

        data format: {
            type: 1,             // meta version
            seed: "moKy",        // user/group name
            key: "{public key}", // PK = secp256k1(SK);
            fingerprint: "..."   // CT = sign(seed, SK);
        }

        algorithm:
            fingerprint = sign(seed, SK);
    """

    @property
    def type(self) -> int:
        """
        Meta algorithm version

            0x01 - username@address
            0x02 - btc_address
            0x03 - username@btc_address
            0x04 - eth_address
            0x05 - username@eth_address
            ...
        """
        raise NotImplemented

    @property
    def key(self) -> VerifyKey:
        """
        Public key (used for signature)

            RSA
            ECC

        :return: public key
        """
        raise NotImplemented

    @property
    def seed(self) -> Optional[str]:
        """
        Seed to generate fingerprint

        :return: ID.name
        """
        raise NotImplemented

    @property
    def fingerprint(self) -> Optional[bytes]:
        """
        Fingerprint to verify ID and public key

            Build: fingerprint = sign(seed, privateKey)
            Check: verify(seed, fingerprint, publicKey)

        :return: signature
        """
        raise NotImplemented

    @abstractmethod
    def generate_address(self, network: int) -> Optional[Address]:
        """
        Generate Address with network type

        :param network:  Address.network
        :return: Address
        """
        raise NotImplemented

    @classmethod
    def check(cls, meta) -> bool:
        key = meta.key
        # meta.key should not be empty
        if isinstance(key, VerifyKey):
            if meta_has_seed(version=meta.type):
                # check seed with signature
                seed = meta.seed
                fingerprint = meta.fingerprint
                # seed and fingerprint should not be empty
                if seed is not None and fingerprint is not None:
                    # verify fingerprint
                    return key.verify(data=utf8_encode(string=seed), signature=fingerprint)
            else:
                # this meta has no seed, so no signature too
                return True

    @classmethod
    def matches(cls, meta, identifier: ID = None, key: VerifyKey = None) -> bool:
        if identifier is not None:
            """ Check whether meta match with entity ID
                (must call this when received a new meta from network) """
            # check ID.name
            if meta.seed == identifier.name:
                # check ID.address
                address = Address.generate(meta=meta, network=identifier.type)
                return address == identifier.address
        elif key is not None:
            """ Check whether meta match with public key """
            if key == meta.key:
                # NOTICE: ID with BTC/ETH address has no username, so
                #         just compare the key.data to check matching
                return True
            # check with seed & fingerprint
            if meta_has_seed(version=meta.type):
                # check whether keys equal by verifying signature
                seed = utf8_encode(string=meta.seed)
                fingerprint = meta.fingerprint
                return key.verify(data=seed, signature=fingerprint)

    #
    #   Factory methods
    #

    @classmethod
    def generate(cls, version: Union[MetaType, int], key: SignKey, seed: Optional[str] = None):  # -> Optional[Meta]:
        factory = cls.factory(version=version)
        assert isinstance(factory, MetaFactory), 'meta type not support: %d, %s' % (version, factory)
        return factory.generate_meta(key=key, seed=seed)

    @classmethod
    def create(cls, version: Union[MetaType, int], key: VerifyKey,
               seed: Optional[str] = None, fingerprint: Union[bytes, str, None] = None):  # -> Optional[Meta]:
        factory = cls.factory(version=version)
        assert isinstance(factory, MetaFactory), 'meta type not support: %d, %s' % (version, factory)
        return factory.create_meta(key=key, seed=seed, fingerprint=fingerprint)

    @classmethod
    def parse(cls, meta: Any):  # -> Optional[Meta]:
        if meta is None:
            return None
        elif isinstance(meta, cls):
            return meta
        elif isinstance(meta, MapWrapper):
            meta = meta.dictionary
        # assert isinstance(meta, dict), 'meta error: %s' % meta
        version = meta_type(meta=meta)
        factory = cls.factory(version=version)
        if factory is None:
            factory = cls.factory(version=0)  # unknown
            assert factory is not None, 'cannot parse meta: %s' % meta
        assert isinstance(factory, MetaFactory), 'meta factory error: %s' % factory
        return factory.parse_meta(meta=meta)

    @classmethod
    def factory(cls, version: Union[MetaType, int]):  # -> Optional[MetaFactory]:
        if isinstance(version, MetaType):
            version = version.value
        return Factories.meta_factories.get(version)

    @classmethod
    def register(cls, version: Union[MetaType, int], factory):
        if isinstance(version, MetaType):
            version = version.value
        Factories.meta_factories[version] = factory


def meta_type(meta: dict) -> int:
    """ get meta type(version) """
    version = meta.get('type')
    if version is None:
        version = meta.get('version')
    return int(version)


class MetaFactory(ABC):

    @abstractmethod
    def generate_meta(self, key: SignKey, seed: Optional[str]) -> Meta:
        """
        Generate meta

        :param key:  private key
        :param seed: ID.name
        :return: Meta
        """
        raise NotImplemented

    @abstractmethod
    def create_meta(self, key: VerifyKey, seed: Optional[str], fingerprint: Union[bytes, str, None]) -> Meta:
        """
        Create meta

        :param key:         public key
        :param seed:        ID.name
        :param fingerprint: sKey.sign(seed)
        :return: Meta
        """
        raise NotImplemented

    @abstractmethod
    def parse_meta(self, meta: dict) -> Optional[Meta]:
        """
        Parse map object to meta

        :param meta: meta info
        :return: Meta
        """
        raise NotImplemented



# -*- coding: utf-8 -*-
#
#   Ming-Ke-Ming : Decentralized User Identity Authentication
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from abc import ABC
from typing import Optional, Union

from .wrappers import Dictionary

from .crypto import base64_encode, base64_decode
from .crypto import VerifyKey, PublicKey

from .types import MetaType, meta_has_seed
from .meta import Meta
from .meta import meta_type


class BaseMeta(Dictionary, Meta, ABC):

    def __init__(self, meta: Optional[dict] = None,
                 version: Union[MetaType, int] = 0, key: Optional[VerifyKey] = None,
                 seed: Optional[str] = None, fingerprint: Union[bytes, str, None] = None):
        # check parameters
        if isinstance(version, MetaType):
            version = version.value
        if fingerprint is None:
            base64 = None
        elif isinstance(fingerprint, bytes):
            base64 = base64_encode(data=fingerprint)
        else:
            assert isinstance(fingerprint, str), 'meta.fingerprint error: %s' % fingerprint
            base64 = fingerprint
            fingerprint = base64_decode(string=base64)
        if meta is None:
            assert version > 0 and key is not None, 'meta error: %d, %s, %s, %s' % (version, key, seed, fingerprint)
            # build meta info
            if seed is None or base64 is None:
                meta = {
                    'type': version,
                    'key': key.dictionary,
                }
            else:
                meta = {
                    'type': version,
                    'key': key.dictionary,
                    'seed': seed,
                    'fingerprint': base64,
                }
        # initialize with meta info
        super().__init__(dictionary=meta)
        # lazy load
        self.__type = version
        self.__key = key
        self.__seed = seed
        self.__fingerprint = fingerprint

    @property  # Override
    def type(self) -> int:
        if self.__type == 0:
            self.__type = meta_type(meta=self.dictionary)
        return self.__type

    @property  # Override
    def key(self) -> VerifyKey:
        if self.__key is None:
            key = self.get('key')
            self.__key = PublicKey.parse(key=key)
        return self.__key

    @property  # Override
    def seed(self) -> Optional[str]:
        if self.__seed is None and meta_has_seed(version=self.type):
            self.__seed = self.get('seed')
        return self.__seed

    @property  # Override
    def fingerprint(self) -> Optional[bytes]:
        if self.__fingerprint is None and meta_has_seed(version=self.type):
            fingerprint = self.get('fingerprint')
            if fingerprint is not None:
                self.__fingerprint = base64_decode(string=fingerprint)
        return self.__fingerprint



# -*- coding: utf-8 -*-
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from abc import ABC, abstractmethod
from typing import Optional, Any

from ..wrappers import MapWrapper

from .cryptography import EncryptKey, DecryptKey, key_algorithm
from .factories import Factories


class SymmetricKey(EncryptKey, DecryptKey, ABC):
    """This class is used to encrypt or decrypt message data

        Symmetric Cryptography Key
        ~~~~~~~~~~~~~~~~~~~~~~~~~~

        key data format: {
            algorithm : "AES", // "DES", ...
            data      : "{BASE64_ENCODE}",
            ...
        }
    """

    AES = 'AES'
    DES = 'DES'

    def __eq__(self, other) -> bool:
        if self is other:
            return True
        if isinstance(other, SymmetricKey):
            return self.match(key=other)

    #
    #  Factory methods
    #

    @classmethod
    def generate(cls, algorithm: str):  # -> Optional[SymmetricKey]:
        factory = cls.factory(algorithm=algorithm)
        assert isinstance(factory, SymmetricKeyFactory), 'key algorithm not support: %s' % algorithm
        return factory.generate_symmetric_key()

    @classmethod
    def parse(cls, key: Any):  # -> Optional[SymmetricKey]:
        if key is None:
            return None
        elif isinstance(key, cls):
            return key
        elif isinstance(key, MapWrapper):
            key = key.dictionary
        # assert isinstance(key, dict), 'key error: %s' % key
        algorithm = key_algorithm(key=key)
        factory = cls.factory(algorithm=algorithm)
        if factory is None:
            factory = cls.factory(algorithm='*')  # unknown
            assert factory is not None, 'cannot parse key: %s' % key
        assert isinstance(factory, SymmetricKeyFactory), 'key algorithm not support: %s' % algorithm
        return factory.parse_symmetric_key(key=key)

    @classmethod
    def factory(cls, algorithm: str):  # -> Optional[SymmetricKeyFactory]:
        return Factories.symmetric_key_factories.get(algorithm)

    @classmethod
    def register(cls, algorithm: str, factory):
        Factories.symmetric_key_factories[algorithm] = factory


class SymmetricKeyFactory(ABC):

    @abstractmethod
    def generate_symmetric_key(self) -> Optional[SymmetricKey]:
        """
        Generate key

        :return: SymmetricKey
        """
        raise NotImplemented

    @abstractmethod
    def parse_symmetric_key(self, key: dict) -> Optional[SymmetricKey]:
        """
        Parse map object to key

        :param key: key info
        :return: SymmetricKey
        """
        raise NotImplemented
